from base import InterfaceObject
from pyticketswitch.util import (
    to_int_or_none, resolve_boolean
)
import order as order_objs
import bundle as bundle_objs
import reservation as res_objs


class Trolley(InterfaceObject):
    """Object that represents a TSW trolley.

    Operations and information relating to a TSW order are available
    through this object, e.g. adding/removing orders, creating a reservation.
    An empty order can be created by calling the constructor with no arguments,
    otherwise the Id of the Trolley can be passed to access an existing TSW
    trolley.

    Args:
        trolley_id (string): Id of the Trolley
        settings (kwargs): See Core constructor.
    """

    def __init__(
        self,
        trolley_id=None,
        core_trolley=None,
        **settings
    ):
        self.trolley_id = trolley_id
        self._core_trolley_attr = core_trolley
        self._orders = None
        self._bundles = None

        super(Trolley, self).__init__(**settings)

    def add_order(self, order):
        """Add the Order object to this Trolley.

        Args:
            order (Order): The Order object to add.
        """
        crypto_block = self.get_crypto_block(
            method_name='start_session',
        )

        resp_dict = self.get_core_api().trolley_add_order(
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            order_token=order.order_id,
            describe_trolley=True,
            trolley_token=self.trolley_id
        )

        self._set_crypto_block(
            crypto_block=resp_dict['crypto_block'],
            method_name='start_session'
        )

        self.trolley_id = resp_dict['trolley_token']
        self._core_trolley = resp_dict['trolley']

    @property
    def order_count(self):
        """Number of orders in this Trolley."""
        return to_int_or_none(
            self._core_trolley.trolley_order_count
        )

    @property
    def is_purchase_successful(self):
        """Check if the trolley was purchased successfully

        Will return True if the purchase was successful,
        False if the purchase was not successful and
        None if no purchase was attempted.
        """
        is_success = None

        if self._core_trolley.purchase_result:

            is_success = resolve_boolean(
                self._core_trolley.purchase_result.success
            )

        return is_success

    @property
    def purchase_failure_reason(self):
        """Reason for purchase failure, None if not attempted or successful."""
        reason = None

        if self.is_purchase_successful is False:
            reason = self._core_trolley.purchase_result.failure_reason

        return reason

    @property
    def purchase_error(self):
        """Error information for failed purchases."""
        purchase_error = None

        if self._core_trolley.purchase_error:
            purchase_error = self._core_trolley.purchase_error

        return purchase_error

    @property
    def _core_trolley(self):
        if self._core_trolley_attr is None:
            self.update()

        if self._core_trolley_attr:
            return self._core_trolley_attr
        else:
            return None

    @_core_trolley.setter
    def _core_trolley(self, value):
        if value is None:
            self._core_trolley_attr = False
        else:
            self._core_trolley_attr = value

        self.orders = None
        self.bundles = None

    def update(self):
        """Updates Trolley data.

        This method is used internally and it shouldn't be necessary
        to call it explicitly.
        """

        crypto_block = self.get_crypto_block(
            method_name='start_session'
        )

        resp_dict = self.get_core_api().trolley_describe(
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            trolley_token=self.trolley_id
        )

        self._core_trolley = resp_dict['trolley']

    def remove_order(self, order_item_number):
        """Remove an Order from this Trolley.

        Args:
            order_item_number (int): The item number of the Order to remove.
        """
        crypto_block = self.get_crypto_block(
            method_name='start_session'
        )

        resp_dict = self.get_core_api().trolley_remove(
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            trolley_token=self.trolley_id,
            remove_item=order_item_number,
            describe_trolley=True
        )

        self._set_crypto_block(
            crypto_block=resp_dict['crypto_block'],
            method_name='start_session'
        )

        self.trolley_id = resp_dict['trolley_token']
        self._core_trolley = resp_dict['trolley']

    def get_reservation(self):
        """Get a Reservation object for this Trolley.

        This method attempts to reserve the Trolley in the TSW system.

        Returns:
            Reservation: Reservation object for this Trolley.
        """
        crypto_block = self.get_crypto_block(
            method_name='start_session'
        )

        resp_dict = self.get_core_api().make_reservation(
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            trolley_token=self.trolley_id,
            self_print_mode='html',
            describe_trolley=True
        )

        if resp_dict['failed_orders']:
            failed_orders = []

            for failed_order in resp_dict['failed_orders']:
                failed_orders.append(
                    order_objs.Order(
                        order_id=failed_order.order_token,
                        core_order=failed_order,
                        **self._internal_settings()
                    )
                )

        else:
            failed_orders = None

        reservation = res_objs.Reservation(
            transaction_id=resp_dict['transaction_id'],
            minutes_left_float=resp_dict['minutes_left_on_reserve'],
            failed_orders=failed_orders,
            need_payment_card=resp_dict['need_payment_card'],
            acceptable_cards=resp_dict.get('acceptable_cards'),
            needs_email_address=resp_dict.get('needs_email_address'),
            supports_billing_address=resp_dict.get('supports_billing_address'),
            needs_agent_reference=resp_dict['needs_agent_reference'],
            prefilled_customer_data=resp_dict.get('prefilled_customer_data'),
            trolley_id=resp_dict['trolley_token'],
            core_trolley=resp_dict['trolley'],
            **self._internal_settings()
        )

        self._set_crypto_for_object(
            crypto_block=resp_dict['crypto_block'],
            method_name='make_reservation',
            interface_object=reservation
        )

        return reservation

    @property
    def orders(self):
        """List of Order objects in this Trolley."""
        if self._orders is None:

            orders = []

            for bundle in self.bundles:

                for order in bundle.orders:

                    orders.append(order)

                self._orders = orders

        return self._orders

    @orders.setter
    def orders(self, value):
        self._orders = value

    def _create_bundle(self, core_bundle):
        return bundle_objs.Bundle(
            core_bundle=core_bundle,
            **self._internal_settings()
        )

    @property
    def bundles(self):
        """List of Bundle objects in this Trolley."""
        if self._bundles is None:

            bundles = []

            for core_bundle in self._core_trolley.bundles:

                bundle = self._create_bundle(
                    core_bundle=core_bundle
                )

                bundles.append(bundle)

            self._bundles = bundles

        return self._bundles

    @bundles.setter
    def bundles(self, value):
        self._bundles = value



import trolley as trolley_objs
from pyticketswitch.util import (
    to_float_or_zero, resolve_boolean, dict_ignore_nones,
    boolean_to_yes_no
)
from base import Customer
import bundle as bundle_objs


class Reservation(trolley_objs.Trolley):
    """Object that represents a reserved TSW trolley.

    In TSW, a reservation is not really a separate element to a trolley,
    but they have some different behaviour and attributes, so in this
    library they have been separated to avoid confusuon. The Reservation
    object is a sub class of the Trolley object.

    Operations and information relating to reservations are accessible
    through this object, e.g. making purchases, reservation details,
    cancelling a reservation. Can be constructed with a transaction Id (
    other constructor arguments are for internal use).

    Args:
        transaction_id (string): Id of the TSW transaction.
        settings (kwargs): See Core constructor.
    """

    def __init__(
        self,
        transaction_id=None,
        minutes_left_float=None,
        failed_orders=None,
        need_payment_card=None,
        acceptable_cards=None,
        needs_email_address=None,
        supports_billing_address=None,
        needs_agent_reference=None,
        prefilled_customer_data=None,
        trolley_id=None,
        core_trolley=None,
        **settings
    ):
        self.transaction_id = transaction_id
        self._minutes_left = None
        self._seconds_left = None
        if minutes_left_float:
            self._set_time_left(minutes_left_float)
        if failed_orders is None:
            failed_orders = []
        self.failed_orders = failed_orders
        self._need_payment_card = need_payment_card
        self._needs_email_address = needs_email_address
        self._supports_billing_address = supports_billing_address
        self._needs_agent_reference = needs_agent_reference
        if prefilled_customer_data is None:
            prefilled_customer_data = {}
        self.prefilled_customer_data = prefilled_customer_data
        self._transaction_status = None
        self._remote_site = None
        self._customer = None
        self._confirmation_page_html = None
        self._self_print_urls = None

        self.acceptable_cards = []

        if acceptable_cards is not None:
            if isinstance(acceptable_cards, dict):
                # The can be changed if _get_acceptable_cards in
                # booking.views is removed
                if isinstance(acceptable_cards['card'], list):
                    # Dealing with acceptable cards created by
                    # _get_acceptable_cards
                    for c in acceptable_cards['card']:

                        self.acceptable_cards.append({
                            'card_type': c['card_type'],
                            'card_desc': c['card_desc']
                        })

                elif isinstance(acceptable_cards['card'], dict):
                    # Dealing with acceptable cards from parse
                    self.acceptable_cards.append({
                        'type': acceptable_cards['card']['card_type'],
                        'description': acceptable_cards['card']['card_desc']
                    })
            else:
                self.acceptable_cards = acceptable_cards

        super(Reservation, self).__init__(
            trolley_id=trolley_id,
            core_trolley=core_trolley,
            **settings
        )

    def _get_cache_key(self):
        return self.transaction_id

    @property
    def _core_trolley(self):
        if self._core_trolley_attr is None:
            self.get_details()

        if self._core_trolley_attr:
            return self._core_trolley_attr
        else:
            return None

    @_core_trolley.setter
    def _core_trolley(self, value):
        if value is None:
            self._core_trolley_attr = False
        else:
            self._core_trolley_attr = value

        self.orders = None
        self.bundles = None

    @property
    def minutes_left(self):
        """Interger number of minutes this reservation is still valid."""
        if self._minutes_left is None:
            self.get_details()

        if self._minutes_left is False:
            return None
        else:
            return self._minutes_left

    @minutes_left.setter
    def minutes_left(self, value):
        if value is None:
            self._minutes_left = False
        else:
            self._minutes_left = value

    @property
    def seconds_left(self):
        """Interger number of seconds in addition to 'minutes_left' that
        this reservation is still valid.
        """
        if self._seconds_left is None:
            self.get_details()

        if self._seconds_left is False:
            return None
        else:
            return self._seconds_left

    @seconds_left.setter
    def seconds_left(self, value):
        if value is None:
            self._seconds_left = False
        else:
            self._seconds_left = value

    def _set_time_left(self, minutes_left_string):

        self.minutes_left = None
        self.seconds_left = None

        if minutes_left_string is not None:
            minutes_float = to_float_or_zero(minutes_left_string)

            if minutes_float > 0:
                self.minutes_left = int(minutes_float)
                self.seconds_left = int(
                    (minutes_float - self.minutes_left) * 60
                )

            else:
                self.minutes_left = 0
                self.seconds_left = 0

    def delete(self):
        """Cancels this reservation, returns True if successful."""

        crypto_block = self._get_crypto_block_for_object(
            method_name='make_reservation',
            interface_object=self
        )

        resp_dict = self.get_core_api().release_reservation(
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
        )

        return resolve_boolean(
            resp_dict['released_ok']
        )

    def purchase_part_one(
        self, return_token, return_domain, return_path,
        return_with_https, encryption_key, customer=None,
        card=None
    ):
        """First part of the 2 stage purchase process.

        See the XML API documentation for more information about the
        purchase process.

        Args:
            return_token (string): Unique token for this purchase call.
            return_domain (string): The domain for the return URL.
            return_path (string): The path for the return URL.
            return_with_https (boolean): Indicates whether the URL should use
                HTTP or HTTPS.
            encryption_key (string): Encryption key for data.
            customer (Customer): Optional, Customer object representing the
                customer making the purchase.
            card (Card): Optional, Card object representing the payment
                card information.

        Returns:
            Dictionary: A dictionary containing an item called
                'redirect_html_page_data' with a string value of
                HTML that should be passed to the user's browser
                in order to perform the redirect.
        """

        if card:
            card_data = dict_ignore_nones(**card._get_dict())
        else:
            card_data = None

        if customer:
            customer_data = dict_ignore_nones(**customer._get_dict())
        else:
            customer_data = None

        https_string = boolean_to_yes_no(
            return_with_https
        )

        crypto_block = self._get_crypto_block_for_object(
            method_name='make_reservation',
            interface_object=self
        )

        resp_dict = self.get_core_api().purchase_reservation_part_one(
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            customer_data=customer_data,
            return_token=return_token, return_domain=return_domain,
            return_path=return_path, return_with_https=https_string,
            encryption_key=encryption_key, card_data=card_data,
        )

        return resp_dict

    def purchase_part_two(
        self, returning_token, new_return_token, new_return_path, http_referer,
        http_accept, http_user_agent, callback_data, encryption_key,
        send_confirmation_email=None, results_url=None
    ):
        """Second part of the 2 stage purchase process.

        See the XML API documentation for more information about the
        purchase process.

        Args:
            returning_token (string): The return token from part one.
            new_return_token (string): A new unique token for this call.
            new_return_path (string): The path for the return URL.
            http_referer (string): The user's 'Referer' HTTP header.
            http_accept (string): The user's 'Accept' HTTP header.
            http_user_agent (string): The user's 'User-Agent' HTTP header.
            callback_data (dictionary): All POST and GET variables.
            encryption_key (string): Encryption key for data.
            send_confirmation_email (boolean): Optional, boolean indicating
                whether TSW should send a confirmation email or not.
            results_url (string): Optional, URL of the confirmation page.

        Returns:
            Dictionary: If an additional redirect is required, then the
                dictionary will contain an item called
                'redirect_html_page_data' which is the same as described in
                part one. This redirect must be followed to complete the
                purchase.
        """

        crypto_block = self.get_crypto_block(
            method_name='start_session',
            password_required=False
        )

        resp_dict = self.get_core_api().purchase_reservation_part_two(
            returning_token=returning_token, new_return_token=new_return_token,
            new_return_path=new_return_path, http_referer=http_referer,
            http_accept=http_accept, http_user_agent=http_user_agent,
            callback_data=callback_data, encryption_key=encryption_key,
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            send_confirmation_email=send_confirmation_email,
            results_url=results_url,
        )

        if 'trolley' in resp_dict:
            self.transaction_id = resp_dict['trolley'].transaction_id
            self._core_trolley = resp_dict['trolley']

        if 'trolley_token' in resp_dict:
            self.trolley_id = resp_dict['trolley_token']

        if 'customer' in resp_dict:
            customer = Customer(
                core_customer=resp_dict['customer']
            )

            resp_dict['customer'] = customer
            self.customer = customer
        else:
            self.customer = None

        if 'self_print_html_pages' in resp_dict:
            self.self_print_urls = resp_dict['self_print_html_pages']

        return resp_dict

    def purchase_reservation(
        self, customer=None, send_confirmation_email=None,
    ):
        """A one stage purchase process that should only be used for purchases
        made on credit.

        See the XML API documentation for more information about the
        purchase process. Note the absence of a 'card_data' element
        in the input - this is intentional since the 2-stage purchase process
        (calling purchase_part_one & purchase_part_two) should always be used
        whenever a payment method is required.

        Args:
            customer (Customer): Optional, Customer object representing the
                customer making the purchase.
            send_confirmation_email (boolean): Optional, boolean indicating
                whether TSW should send a confirmation email or not.

        Returns:
            Dictionary: The result of the attempted purchase
        """

        if customer:
            customer_data = dict_ignore_nones(**customer._get_dict())
        else:
            customer_data = None

        crypto_block = self._get_crypto_block_for_object(
            method_name='make_reservation',
            interface_object=self
        )

        resp_dict = self.get_core_api().purchase_reservation(
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            customer_data=customer_data,
            send_confirmation_email=send_confirmation_email,
        )

        return resp_dict

    @property
    def transaction_status(self):
        """Status of the reservation."""
        if self._transaction_status is None:
            self.get_details()

        if self._transaction_status:
            return self._transaction_status
        else:
            return None

    @transaction_status.setter
    def transaction_status(self, value):
        if value is None:
            self._transaction_status = False
        else:
            self._transaction_status = value

    @property
    def remote_site(self):
        """The domain that the transaction was made on."""
        if self._remote_site is None:
            self.get_details()

        if self._remote_site:
            return self._remote_site
        else:
            return None

    @remote_site.setter
    def remote_site(self, value):
        if value is None:
            self._remote_site = False
        else:
            self._remote_site = value

    @property
    def customer(self):
        """Customer object representing the customer on the reservation."""
        if self._customer is None:
            self.get_details()

        if self._customer:
            return self._customer
        else:
            return None

    @customer.setter
    def customer(self, value):
        if value is None:
            self._customer = False
        else:
            self._customer = value

    @property
    def confirmation_page_html(self):
        """HTML of the confirmation page, if it was stored."""
        if self._confirmation_page_html is None:
            self.get_details()

        if self._confirmation_page_html:
            return self._confirmation_page_html
        else:
            return None

    @confirmation_page_html.setter
    def confirmation_page_html(self, value):
        if value is None:
            self._confirmation_page_html = False
        else:
            self._confirmation_page_html = value

    def get_details(self):
        """Retrieves information about the reservation.

        Called internally by several methods/properties, it shouldn't
        be necessary to call this method explicitly.

        Returns:
            Reservation: Self, with updated data.
        """
        crypto_block = self.get_crypto_block(
            method_name='start_session',
            password_required=False
        )

        resp_dict = self.get_core_api().transaction_info(
            transaction_id=self.transaction_id, describe_trolley=True,
            describe_customer=True, describe_external_sale_page=True,
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
        )

        self.transaction_status = resp_dict['transaction_status']
        self._set_time_left(resp_dict['minutes_left_on_reserve'])
        self.remote_site = resp_dict['remote_site']

        self._core_trolley = resp_dict['trolley']

        if 'customer' in resp_dict:

            lang_str = resp_dict.get('language_list', None)

            if lang_str:
                langs = lang_str.split(',')

            self.customer = Customer(
                core_customer=resp_dict['customer'],
                languages=langs,
            )

        else:
            self.customer = None

        if 'sale_page' in resp_dict:
            self.confirmation_page_html = resp_dict['sale_page'].sale_page
        else:
            self.confirmation_page_html = None

        if 'self_print_html_pages' in resp_dict:
            self.self_print_urls = resp_dict['self_print_html_pages']

        return self

    @property
    def self_print_urls(self):
        """List of URLs for self print vouchers on this reservation."""
        if self._self_print_urls is None:
            self.get_details()

        return self._self_print_urls

    @self_print_urls.setter
    def self_print_urls(self, value):
        self._self_print_urls = value

    def set_confirmation_page(self, html):
        """Store the confirmation page HTML in the TSW system.

        Args:
            html (string): HTML to store.

        Returns:
            Boolean: True if data was saved successfully.
        """

        crypto_block = self.get_crypto_block(
            method_name='start_session',
            password_required=False
        )

        resp_dict = self.get_core_api().save_external_sale_page(
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            transaction_id=self.transaction_id,
            sale_page_type='text', sale_page_subtype='html', sale_page=html,
        )

        if resp_dict['saved_this_page'] == 'yes':
            return True
        else:
            return False

    @property
    def is_released(self):
        """Boolean indicating if the reservation was released."""
        if self.transaction_status == 'released':
            return True
        elif self.transaction_status:
            return False
        return None

    @property
    def is_purchased(self):
        """Boolean indicating if the reservation was purchased successfully."""
        if self.transaction_status == 'purchased':
            return True
        elif self.transaction_status:
            return False
        return None

    @property
    def is_purchase_successful(self):
        """Boolean indicating if the reservation was purchased successfully.

        Overrides trolley function.
        """
        return self.is_purchased

    @property
    def is_reserved(self):
        """Boolean indicating if the Reservation object is reserved."""
        if self.transaction_status == 'reserved':
            return True
        elif self.transaction_status:
            return False
        return None

    @property
    def is_attempting(self):
        """Boolean indicating if TSW is currently attempting to reserve."""
        if self.transaction_status == 'attempting':
            return True
        elif self.transaction_status:
            return False
        return None

    def _create_bundle(self, core_bundle):
        return bundle_objs.Bundle(
            core_bundle=core_bundle,
            self_print_urls=self.self_print_urls,
            **self._internal_settings()
        )

    @property
    def need_payment_card(self):
        """Boolean indicating if a payment card is required for purchase."""
        return resolve_boolean(
            self._need_payment_card
        )

    @property
    def needs_email_address(self):
        """Boolean indicating if an email address is required for purchase."""
        return resolve_boolean(
            self._needs_email_address
        )

    @property
    def supports_billing_address(self):
        """Boolean indicating if billing address can be provided during
        purchase."""
        return resolve_boolean(
            self._supports_billing_address
        )

    @property
    def needs_agent_reference(self):
        return resolve_boolean(
            self._needs_agent_reference
        )



from base import InterfaceObject
from pyticketswitch import settings
import event as event_objs
import order as order_objs
import reservation as res_objs
from pyticketswitch.util import date_to_yyyymmdd


class Core(InterfaceObject):
    """Object that represents the core API functionality

    Operations that are not related to a particular object are accessible
    from this object, such as the event search. As such, it is commonly the
    starting point for API interactions.

    The only arguments are configuration settings, which are described below
    and are required when constructing several other objects.

    Args:
        username (string): TSW user
        password (string): password for TSW user
        url (string): TSW API URL
        accept_language (string): user's HTTP Accept-Language header
        api_request_timeout (int): Optional, API timeout in seconds
        no_time_descr (string): Optional, text to use if no time is returned
            by the API for a performance
        default_concession_descr (string): Optional, text to use if no
            description is returned by the API for a concession
    """

    def __init__(
        self, **settings
    ):
        self._setup_instance_variables()

        super(Core, self).__init__(**settings)

    def _setup_instance_variables(self):

        self.events = []
        self._cities = {}
        self._countries = {}
        self._categories = {}
        self._custom_fields = {}
        self._custom_filters = {}
        self._min_seatprice_range = []

    def _do_core_event_search(
            self, crypto_block, upfront_data_token, s_keys, s_dates, s_coco,
            s_city, s_geo_lat, s_geo_long, s_geo_rad_km,
            s_src, s_area, s_ven, s_eve,
            s_class, event_token_list,
            request_source_info, request_extra_info, request_video_iframe,
            request_cost_range, request_media, request_custom_fields,
            request_reviews, request_avail_details,
            s_top, s_user_rating, s_critic_rating,
            s_auto_range, page_length, page_number,
            s_cust_fltr, s_airport, mime_text_type,
            special_offer_only, events=None, iter_index=0, max_iterations=None):

        # There is no filter in the core for special offers, so if only
        # the special offers are requested, then we need to recursively
        # call the event search until we have enough special offer events
        # (or there aren't any more events returned). By default, we do 3 small
        # searches followed by a full search. If max_iterations is provided, we
        # do that many small searches only.
        #
        # If the special_offer_only flag is False, then we can do a search
        # as normal
        if special_offer_only:

            if events is None:
                events = {}

            if not page_length:

                num_required = 0
                num_to_request = None
            else:

                if page_number:
                    num_required = page_length * (page_number + 1)
                else:
                    num_required = page_length

                # if the event search has been called multiple times
                # already, then do a full search
                if not max_iterations and iter_index > 2:
                    num_to_request = None
                else:
                    # arbitrarily choose to request twice as many events
                    # as required looking for special offers
                    num_to_request = num_required * 2

            resp_dict = self.get_core_api().event_search(
                crypto_block=crypto_block,
                upfront_data_token=upfront_data_token,
                s_keys=s_keys, s_dates=s_dates,
                s_coco=s_coco, s_city=s_city,
                s_geo_lat=s_geo_lat, s_geo_long=s_geo_long,
                s_geo_rad_km=s_geo_rad_km, s_src=s_src, s_area=s_area,
                s_ven=s_ven, s_eve=s_eve, s_class=s_class,
                event_token_list=event_token_list,
                request_source_info=request_source_info,
                request_extra_info=request_extra_info,
                request_video_iframe=request_video_iframe,
                request_cost_range=request_cost_range,
                request_media=request_media,
                request_custom_fields=request_custom_fields,
                request_reviews=request_reviews,
                request_avail_details=request_avail_details,
                s_top=s_top, s_user_rating=s_user_rating,
                s_critic_rating=s_critic_rating, s_auto_range=s_auto_range,
                page_length=num_to_request, page_number=iter_index,
                s_cust_fltr=s_cust_fltr, s_airport=s_airport,
                mime_text_type=mime_text_type,
            )

            # If the event has a special offer, then add it to the list
            for e in resp_dict['event']:
                if e.cost_range and (
                    e.cost_range.best_value_offer or
                    e.cost_range.max_saving_offer or
                    e.cost_range.top_price_offer
                ):

                    if e.event_id not in events:
                        events[e.event_id] = e

            # Return the list of special offer events if we've retrieved the
            # number required or there aren't any more or if max_iterations
            # has been reached
            if (
                not resp_dict['event'] or
                len(events) >= num_required or
                num_to_request is None or
                len(resp_dict['event']) < num_to_request or
                (max_iterations and (iter_index + 1) == max_iterations)
            ):

                if page_length is None:
                    resp_dict['event'] = events.values()

                else:

                    start = (page_number or 0) * page_length
                    end = ((page_number or 0) + 1) * page_length

                    resp_dict['event'] = events.values()[start:end]

                return resp_dict

            # Call recursively if we haven't found enough special offer
            # events yet
            else:

                iter_index = iter_index + 1

                return self._do_core_event_search(
                    crypto_block=crypto_block,
                    upfront_data_token=upfront_data_token,
                    s_keys=s_keys, s_dates=s_dates, s_coco=s_coco,
                    s_city=s_city, s_geo_lat=s_geo_lat,
                    s_geo_long=s_geo_long, s_geo_rad_km=s_geo_rad_km,
                    s_src=s_src, s_area=s_area, s_ven=s_ven,
                    s_eve=s_eve, s_class=s_class,
                    event_token_list=event_token_list,
                    request_source_info=request_source_info,
                    request_extra_info=request_extra_info,
                    request_video_iframe=request_video_iframe,
                    request_cost_range=request_cost_range,
                    request_media=request_media,
                    request_custom_fields=request_custom_fields,
                    request_reviews=request_reviews,
                    request_avail_details=request_avail_details,
                    s_top=s_top, s_user_rating=s_user_rating,
                    s_critic_rating=s_critic_rating,
                    s_auto_range=s_auto_range, page_length=page_length,
                    page_number=page_number,
                    s_cust_fltr=s_cust_fltr, s_airport=s_airport,
                    special_offer_only=special_offer_only,
                    events=events, iter_index=iter_index,
                    mime_text_type=mime_text_type,
                    max_iterations=max_iterations,
                )

        else:
            return self.get_core_api().event_search(
                crypto_block=crypto_block,
                upfront_data_token=upfront_data_token,
                s_keys=s_keys, s_dates=s_dates,
                s_coco=s_coco, s_city=s_city,
                s_geo_lat=s_geo_lat, s_geo_long=s_geo_long,
                s_geo_rad_km=s_geo_rad_km, s_src=s_src, s_area=s_area,
                s_ven=s_ven, s_eve=s_eve, s_class=s_class,
                event_token_list=event_token_list,
                request_source_info=request_source_info,
                request_extra_info=request_extra_info,
                request_video_iframe=request_video_iframe,
                request_cost_range=request_cost_range,
                request_media=request_media,
                request_custom_fields=request_custom_fields,
                request_reviews=request_reviews,
                request_avail_details=request_avail_details,
                s_top=s_top, s_user_rating=s_user_rating,
                s_critic_rating=s_critic_rating, s_auto_range=s_auto_range,
                page_length=page_length, page_number=page_number,
                s_cust_fltr=s_cust_fltr, s_airport=s_airport,
                mime_text_type=mime_text_type,
            )

    def search_events(
            self, keyword=None,
            earliest_date=None, latest_date=None,
            country=None, city=None,
            latitude=None, longitude=None, radius=None,
            source=None, area_code=None, venue_code=None,
            event_code=None, category=None, event_id_list=None,
            page_length=None, page_number=None,
            sort_by=None, auto_date_range=None,
            request_source_info=None, request_extra_info=None,
            request_video_iframe=None, request_cost_range=True,
            request_media=None, request_custom_fields=True,
            request_reviews=None, request_avail_details=None,
            custom_filter_list=None, airport=None, special_offer_only=False,
            mime_text_type=None, max_iterations=None):
        """Perform event search, returns list of Event objects.

        If no arguments are provided, then the full list of Events
        that are available will be returned.

        Args:
            keyword (string): keyword search key (default None)
            earliest_date (datetime): restrict to Events with Performances
                after this date (default None)
            latest_date (datetime): restrict to Events with Performances
                before this date (default None)
            country (string): ISO 3166-1 country code (default None)
            city (string): TSW city key (default None)
            latitude (float/int/string): (default None)
            longitude (float/int/string): (default None)
            radius (int/string): integer radius from lat/long point in
                km (default None)
            source (string): supplier system name (default None)
            area_code (string): TSW area code (default None)
            venue_code (string): TSW venue code (default None)
            event_code (string): TSW event code (default None)
            category (string): Event category search key (default None)
            event_id_list (list): Event Ids to return (default None)
            page_length (int): for pagination, number of Event objects
                to return (default None)
            page_number (int): for pagination, page number (default None)
            sort_by (string): defines sorting order, either top (best selling),
                user_rating or critic_rating (default None)
            auto_date_range (string): see settings.AUTO_DATE_RANGE for options
                (default None)
            request_source_info (boolean): flag for supplier info
                (default None)
            request_extra_info (boolean): flag for extra info (default None)
            request_video_iframe (boolean): flag for video (default None)
            request_cost_range (boolean): flag for cost range data
                (default True)
            request_media (list): List of strings representing the names
                of the images to request (default None)
            request_custom_fields (boolean): flag for custom fields
                (default True)
            request_reviews (boolean): flag for reviews (default None)
            request_avail_details (boolean): flag for avail_details
                (default None)
            custom_filter_list (list): custom filters to apply (default None)
            airport (string): airport code (default None)
            special_offer_only (boolean): flag for only returning events with
                special offers (default False)
            mime_text_type (string): desired text format for certain fields
                (most common options are 'html' and 'plain') (default None)
            max_iterations (int): used only in conjunction with
                special_offer_only. Sets the maximum number of iterations
                when retrieving special offers to prevent a full product search
                being performed.

        Returns:
            list: List of Event objects
        """

        s_top = None
        s_user_rating = None
        s_critic_rating = None

        if sort_by:
            if sort_by == 'top':
                s_top = True
            elif sort_by == 'user_rating':
                s_user_rating = True
            elif sort_by == 'critic_rating':
                s_critic_rating = True

        if earliest_date or latest_date:

            date_range = ['', '']
            if earliest_date:
                date_range[0] = date_to_yyyymmdd(earliest_date)
            if latest_date:
                date_range[1] = date_to_yyyymmdd(latest_date)

            date_range = ':'.join(date_range)
        else:
            date_range = None

        if auto_date_range in settings.AUTO_DATE_RANGE:
            s_auto_range = auto_date_range
        else:
            s_auto_range = None

        if request_media is None:
            request_media = settings.REQUEST_MEDIA

        if custom_filter_list:
            s_cust_fltr = ' '.join(custom_filter_list)
        else:
            s_cust_fltr = None

        if event_id_list:
            event_token_list = ','.join(event_id_list)
        else:
            event_token_list = None

        crypto_block = self.get_crypto_block(
            method_name='start_session',
            password_required=False
        )

        if self.events:
            self._setup_instance_variables()

        resp_dict = self._do_core_event_search(
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            s_keys=keyword, s_dates=date_range,
            s_coco=country, s_city=city,
            s_geo_lat=latitude, s_geo_long=longitude,
            s_geo_rad_km=radius, s_src=source, s_area=area_code,
            s_ven=venue_code, s_eve=event_code, s_class=category,
            event_token_list=event_token_list,
            request_source_info=request_source_info,
            request_extra_info=request_extra_info,
            request_video_iframe=request_video_iframe,
            request_cost_range=request_cost_range,
            request_media=request_media,
            request_custom_fields=request_custom_fields,
            request_reviews=request_reviews,
            request_avail_details=request_avail_details,
            s_top=s_top, s_user_rating=s_user_rating,
            s_critic_rating=s_critic_rating, s_auto_range=s_auto_range,
            page_length=page_length, page_number=page_number,
            s_cust_fltr=s_cust_fltr, s_airport=airport,
            special_offer_only=special_offer_only,
            mime_text_type=mime_text_type, max_iterations=max_iterations
        )

        requested_data = {}

        if request_source_info:
            requested_data['source_info'] = True

        if request_extra_info:
            requested_data['extra_info'] = True

        if request_video_iframe:
            requested_data['video_iframe'] = True

        if request_cost_range:
            requested_data['cost_range'] = True

        if request_custom_fields:
            requested_data['custom_fields'] = True

        if request_reviews:
            requested_data['reviews'] = True

        if request_avail_details:
            requested_data['avail_details'] = True

        if request_media:
            requested_data['media'] = {}

            for m in request_media:
                requested_data['media'][m] = True

        events = []

        for core_event in resp_dict['event']:
            # create event objects and append to list
            event = event_objs.Event(
                event_id=core_event.event_token,
                core_event=core_event,
                requested_data=requested_data,
                **self._internal_settings()
            )
            events.append(event)

            if event.city_code:
                self._add_city(event.city_code, event.city_desc)

            if event.country_code:
                self._add_country(event.country_code, event.country_desc)

            if event.categories:
                self._add_categories(event.categories)

            if request_cost_range and event.min_seatprice_float:
                self._min_seatprice_range.append(event.min_seatprice_float)

            if request_custom_fields and event.custom_fields:
                self._add_custom_fields(event.custom_fields)

            if event.custom_filters:
                self._add_custom_filters(event.custom_filters)

        self._set_crypto_block(
            crypto_block=resp_dict['crypto_block'],
            method_name='event_search'
        )

        self.events = events

        return events

    def _add_city(self, code, desc):
        if code in self._cities:
            self._cities[code]['count'] += 1

        else:
            self._cities[code] = {'description': desc, 'count': 1}

    def _add_country(self, code, desc):
        if code in self._countries:
            self._countries[code]['count'] += 1

        else:
            self._countries[code] = {'description': desc, 'count': 1}

    @staticmethod
    def _add_cat_to_dict(category, category_dict):
        if category.search_key in category_dict:
            category_dict[category.search_key]['count'] += 1

        else:
            category_dict[category.search_key] = {
                'count': 1,
                'category': category,
                'sub_categories': {},
            }

        if category.sub_categories:

            for sub in category.sub_categories:

                Core._add_cat_to_dict(
                    sub,
                    category_dict[category.search_key]['sub_categories']
                )

    def _add_categories(self, categories):

        for cat in categories:

            Core._add_cat_to_dict(
                category=cat,
                category_dict=self._categories
            )

    @staticmethod
    def _add_custom_field_to_dict(custom_field, custom_field_dict):
        if custom_field.code in custom_field_dict:
            custom_field_dict[custom_field.code]['count'] += 1

        else:
            custom_field_dict[custom_field.code] = {
                'count': 1,
                'custom_field': custom_field,
            }

    def _add_custom_fields(self, custom_fields):

        for cf in custom_fields:
            Core._add_custom_field_to_dict(
                custom_field=cf,
                custom_field_dict=self._custom_fields
            )

    @staticmethod
    def _add_custom_filter_to_dict(custom_filter, custom_filter_dict):
        if custom_filter.key in custom_filter_dict:
            custom_filter_dict[custom_filter.key]['count'] += 1

        else:
            custom_filter_dict[custom_filter.key] = {
                'count': 1,
                'custom_filter': custom_filter,
            }

    def _add_custom_filters(self, custom_filters):

        for cf in custom_filters:
            Core._add_custom_filter_to_dict(
                custom_filter=cf,
                custom_filter_dict=self._custom_filters
            )

    @property
    def event_cities(self):
        """Dictionary of cities in the search results.

        A dictionary of cities with the code, name and a count
        of how many times it appeared in the search.
        """
        return self._cities

    @property
    def event_categories(self):
        """Dictionary of Category objects in the search results.

        A dictionary of Category objects by their code, also contains
        the count of how many times it appeared and the sub categories.
        """
        return self._categories

    @property
    def event_countries(self):
        """Dictionary of countries in the search results.

        A dictionary of countries with the code, name and a count
        of how many times it appeared in the search.
        """
        return self._countries

    @property
    def event_price_range(self):
        """List of minimum prices in the search results.

        A list of float values of minimum prices of the Events
        in the search, where the information is available.
        Uses the Event.min_combined_price_float value.
        """
        if not self._min_seatprice_range:
            for e in self.events:
                if e.min_combined_price_float:
                    self._min_seatprice_range.append(
                        e.min_combined_price_float
                    )

        return self._min_seatprice_range

    @property
    def event_custom_fields(self):
        """Dictionary of CustomField objects in the search results.

        A dictionary of CustomField objects by their code, also contains
        the count of how many times it appeared.
        """
        return self._custom_fields

    @property
    def event_custom_filters(self):
        """Dictionary of CustomFilter objects in the search results.

        A dictionary of CustomFilter objects by their code, also contains
        the count of how many times it appeared.
        """
        return self._custom_filters

    def create_order(
            self, concessions=None, despatch_method=None):
        """Create a new Order.

        Order objects can be added to a Trolley, allowing multiple
        items to be purchased.

        Args:
            concessions (list): the list of Concession objects to
                create the Order with.
            despatch_method (DespatchMethod): the selected DespatchMethod
                for the Order.

        Returns:
            Order: An Order object
        """

        crypto_block = None

        if not concessions:
            concession_ids = None

            crypto_block = self.get_crypto_block(
                method_name='discount_options',
            )

        else:
            concession_ids = []
            for c in concessions:
                concession_ids.append(
                    c.concession_id
                )
                if not crypto_block:
                    crypto_block = self._get_crypto_block_for_object(
                        method_name='discount_options',
                        interface_object=c
                    )

        if despatch_method is not None:
            despatch_id = despatch_method.despatch_id
        else:
            despatch_id = None

        resp_dict = self.get_core_api().create_order(
            crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            discount_token=concession_ids, despatch_token=despatch_id
        )

        order = order_objs.Order(
            order_id=resp_dict['order_token'],
            core_order=resp_dict['order'],
            core_currency=resp_dict['currency'],
            **self._internal_settings()
        )

        self._set_crypto_block(
            crypto_block=resp_dict['crypto_block'],
            method_name='start_session'
        )

        return order

    def create_reservation(
            self, concessions=None, despatch_method=None):
        """Create a new Reservation object.

        Creates a Reservation object that contains only a single
        Order containing the passed values. It is equivalent to
        creating an Order, adding this to a Trolley and then
        reserving that Trolley.

        Args:
            concessions (list): the list of Concession objects to
                create the Reservation with.
            despatch_method (DespatchMethod): the selected DespatchMethod
                for the Reservation.

        Returns:
            Reservation: A Reservation object
        """

        crypto_block = None

        if not concessions:
            concession_ids = None

            crypto_block = self.get_crypto_block(
                method_name='discount_options',
            )

        else:
            concession_ids = []
            for c in concessions:
                concession_ids.append(
                    c.concession_id
                )
                if not crypto_block:
                    crypto_block = self._get_crypto_block_for_object(
                        method_name='discount_options',
                        interface_object=c
                    )

        if despatch_method is not None:
            despatch_id = despatch_method.despatch_id
        else:
            despatch_id = None

        resp_dict = self.get_core_api().create_order_and_reserve(
            crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            discount_token=concession_ids, despatch_token=despatch_id
        )

        if resp_dict['failed_orders']:
            failed_orders = []

            for failed_order in resp_dict['failed_orders']:
                failed_orders.append(
                    order_objs.Order(
                        order_id=failed_order.order_token,
                        core_order=failed_order,
                        **self._internal_settings()
                    )
                )

        else:
            failed_orders = None

        reservation = res_objs.Reservation(
            transaction_id=resp_dict['transaction_id'],
            minutes_left_float=resp_dict['minutes_left_on_reserve'],
            failed_orders=failed_orders,
            need_payment_card=resp_dict['need_payment_card'],
            acceptable_cards=resp_dict.get('acceptable_cards'),
            needs_email_address=resp_dict.get('needs_email_address'),
            supports_billing_address=resp_dict.get('supports_billing_address'),
            needs_agent_reference=resp_dict['needs_agent_reference'],
            prefilled_customer_data=resp_dict.get('prefilled_customer_data'),
            trolley_id=resp_dict['trolley_token'],
            core_trolley=resp_dict['trolley'],
            **self._internal_settings()
        )

        self._set_crypto_for_object(
            crypto_block=resp_dict['crypto_block'],
            method_name='make_reservation',
            interface_object=reservation
        )

        return reservation




class CoreObject(object):
    pass


class CoreObjectCollection(list):

    def all(self):
        return self


class RunningUser(CoreObject):

    def __init__(
        self,
        user_id,
        style,
        real_name=None,
        sub_style=None,
        sub_sub_style=None,
        backend_group=None,
        content_group=None,
        restrict_group=None,
        sphinx_restrict_group=None,
        default_country_code=None,
        default_lang_code=None,
        **kwargs
    ):

        self.user_id = user_id
        self.style = style
        self.real_name = real_name
        self.sub_style = sub_style
        self.sub_sub_style = sub_sub_style
        self.backend_group = backend_group
        self.content_group = content_group
        self.restrict_group = restrict_group
        self.sphinx_restrict_group = sphinx_restrict_group
        self.default_country_code = default_country_code
        self.default_lang_code = default_lang_code

        vars(self).update(kwargs)


class Event(CoreObject):

    def __init__(
        self,
        event_desc,
        venue_desc,
        source_desc,
        source_code,
        event_token=None,
        event_id=None,
        venue_info=None,
        venue_addr=None,
        classes=None,
        event_medias=None,
        cost_range=None,
        city_code=None,
        city_desc=None,
        country_code=None,
        country_desc=None,
        geo_data=None,
        reviews=None,
        user_review_percent=None,
        critic_review_percent=None,
        source_t_and_c=None,
        source_after_sales_email=None,
        source_card_statement_desc=None,
        source_enquiries_email=None,
        source_local_phone=None,
        source_local_fax=None,
        source_international_phone=None,
        source_international_fax=None,
        source_postal_addr=None,
        video_iframe=None,
        custom_fields=None,
        custom_filters=None,
        is_seated=None,
        date_range_start=None,
        date_range_end=None,
        show_perf_time=None,
        need_departure_date=None,
        need_performance=None,
        need_duration=None,
        structured_info=None,
        event_quantity_options=None,
        avail_details=None,
        **kwargs
    ):

        self.event_desc = event_desc
        self.venue_desc = venue_desc
        self.source_desc = source_desc
        self.source_code = source_code
        self.event_token = event_token
        self.event_id = event_id

        if event_token is None and event_id is not None:
            self.event_token = event_id

        if self.event_id is None and self.event_token is not None:
            self.event_id = event_token

        self.venue_info = venue_info
        self.venue_addr = venue_addr
        if classes is None:
            classes = []
        self.classes = classes
        if event_medias is None:
            event_medias = []
        self.event_medias = event_medias
        self.cost_range = cost_range
        self.city_code = city_code
        self.city_desc = city_desc
        self.country_code = country_code
        self.country_desc = country_desc
        self.geo_data = geo_data
        if reviews is None:
            reviews = []
        self.reviews = reviews
        self.user_review_percent = user_review_percent
        self.critic_review_percent = critic_review_percent
        self.source_t_and_c = source_t_and_c
        self.source_after_sales_email = source_after_sales_email
        self.source_card_statement_desc = source_card_statement_desc
        self.source_enquiries_email = source_enquiries_email
        self.source_local_phone = source_local_phone
        self.source_local_fax = source_local_fax
        self.source_international_phone = source_international_phone
        self.source_international_fax = source_international_fax
        self.source_postal_addr = source_postal_addr
        self.video_iframe = video_iframe
        if custom_fields is None:
            custom_fields = []
        self.custom_fields = custom_fields
        if custom_filters is None:
            custom_filters = []
        self.custom_filters = custom_filters
        self.is_seated = is_seated
        self.date_range_start = date_range_start
        self.date_range_end = date_range_end
        self.show_perf_time = show_perf_time
        self.need_departure_date = need_departure_date
        self.need_performance = need_performance
        self.need_duration = need_duration
        self.need_duration = need_duration
        if structured_info is None:
            structured_info = {}
        self.structured_info = structured_info
        self.event_quantity_options = event_quantity_options
        self.avail_details = avail_details

        vars(self).update(kwargs)

    def add_extra_info(self, extra_info_event):

        for k, v in vars(extra_info_event).items():
            if v is not None:
                setattr(self, k, v)


class Class(CoreObject):

    def __init__(
        self,
        class_code,
        class_desc=None,
        is_main_class=None,
        search_key=None,
        subclasses=None,
        **kwargs
    ):

        self.class_code = class_code
        self.class_desc = class_desc
        self.is_main_class = is_main_class
        self.search_key = search_key
        if subclasses is None:
            subclasses = []
        self.subclasses = subclasses

        vars(self).update(kwargs)


class SubClass(CoreObject):
    objects = CoreObjectCollection()

    def __init__(
        self,
        subclass_code,
        subclass_desc=None,
        is_main_subclass=None,
        search_key=None,
        **kwargs
    ):
        self.subclass_code = subclass_code
        self.subclass_desc = subclass_desc
        self.is_main_subclass = is_main_subclass
        self.search_key = search_key

        vars(self).update(kwargs)

        self.objects.append(self)


class GeoData(CoreObject):

    def __init__(self, latitude, longitude):
        self.latitude = latitude
        self.longitude = longitude


class EventMedia(CoreObject):

    def __init__(
        self,
        name,
        path,
        host,
        secure_complete_url,
        insecure_complete_url,
        **kwargs
    ):
        self.name = name
        self.path = path
        self.host = host
        self.secure_complete_url = secure_complete_url
        self.insecure_complete_url = insecure_complete_url

        vars(self).update(kwargs)


class VideoIframe(CoreObject):

    def __init__(
        self,
        video_iframe_height,
        video_iframe_width,
        video_iframe_host,
        video_iframe_path,
        video_iframe_supports_https,
        video_iframe_url_when_insecure=None,
        video_iframe_url_when_secure=None,
        **kwargs
    ):

        self.video_iframe_height = video_iframe_height
        self.video_iframe_width = video_iframe_width
        self.video_iframe_host = video_iframe_host
        self.video_iframe_path = video_iframe_path
        self.video_iframe_supports_https = video_iframe_supports_https
        self.video_iframe_url_when_insecure = video_iframe_url_when_insecure
        self.video_iframe_url_when_secure = video_iframe_url_when_secure

        vars(self).update(kwargs)


class CustomField(CoreObject):

    def __init__(
        self,
        custom_field_data=None,
        custom_field_label=None,
        custom_field_name=None,
        **kwargs
    ):
        self.custom_field_data = custom_field_data
        self.custom_field_label = custom_field_label
        self.custom_field_name = custom_field_name

        vars(self).update(kwargs)


class CustomFilter(CoreObject):

    def __init__(
        self,
        custom_filter_desc=None,
        custom_filter_key=None,
        **kwargs
    ):
        self.custom_filter_desc = custom_filter_desc
        self.custom_filter_key = custom_filter_key

        vars(self).update(kwargs)


class Performance(CoreObject):

    def __init__(
        self,
        is_limited,
        perf_token=None,
        date_desc=None,
        date_utc_offset=None,
        date_utc_seconds=None,
        date_yyyymmdd=None,
        has_pool_seats=None,
        perf_is_visible=None,
        perf_subdata=None,
        perf_type_code=None,
        running_time=None,
        time_desc=None,
        time_hhmmss=None,
        perf_name=None,
        cost_range=None,
        cached_max_seats=None,
        **kwargs
    ):
        self.is_limited = is_limited
        self.perf_token = perf_token
        self.date_desc = date_desc
        self.date_utc_offset = date_utc_offset
        self.date_utc_seconds = date_utc_seconds
        self.date_yyyymmdd = date_yyyymmdd
        self.has_pool_seats = has_pool_seats
        self.perf_is_visible = perf_is_visible
        self.perf_subdata = perf_subdata
        self.perf_type_code = perf_type_code
        self.running_time = running_time
        self.time_desc = time_desc
        self.time_hhmmss = time_hhmmss
        self.perf_name = perf_name
        self.cost_range = cost_range
        self.cached_max_seats = cached_max_seats

        vars(self).update(kwargs)


class Month(CoreObject):

    def __init__(
        self,
        year_number,
        month_number,
        short_month_name,
        long_month_name,
        earliest_date,
        latest_date,
        **kwargs
    ):

        self.year_number = year_number
        self.month_number = month_number
        self.short_month_name = short_month_name
        self.long_month_name = long_month_name
        self.earliest_date = earliest_date
        self.latest_date = latest_date

        vars(self).update(kwargs)


class TicketType(CoreObject):

    def __init__(
        self,
        ticket_type_desc,
        price_bands,
        ticket_type_token=None,
        ticket_type_code=None,
        **kwargs
    ):
        self.ticket_type_desc = ticket_type_desc
        self.price_bands = price_bands
        self.ticket_type_token = ticket_type_token
        self.ticket_type_code = ticket_type_code

        vars(self).update(kwargs)


class PriceBand(CoreObject):

    def __init__(
        self,
        ticket_price,
        surcharge,
        number_available,
        is_offer,
        band_token,
        combined=None,
        seatprice=None,
        non_offer_combined=None,
        non_offer_seatprice=None,
        non_offer_surcharge=None,
        non_offer_ticket_price=None,
        percentage_saving=None,
        example_seats=None,
        price_band_code=None,
        price_band_desc=None,
        possible_discounts=None,
        example_seats_are_real=None,
        discount_code=None,
        discount_desc=None,
        discount_subdata=None,
        free_seat_blocks=None,
        raw_contiguous_seats=None,
        raw_total_seats=None,
        user_commission=None,
        gross_commission=None,
        **kwargs
    ):
        self.ticket_price = ticket_price
        self.surcharge = surcharge
        self.number_available = number_available
        self.is_offer = is_offer
        self.band_token = band_token
        self.combined = combined
        self.seatprice = seatprice
        self.non_offer_combined = non_offer_combined
        self.non_offer_seatprice = non_offer_seatprice
        self.non_offer_surcharge = non_offer_surcharge
        self.non_offer_ticket_price = non_offer_ticket_price
        self.percentage_saving = percentage_saving
        if example_seats is None:
            example_seats = []
        self.example_seats = example_seats
        self.price_band_code = price_band_code
        self.price_band_desc = price_band_desc
        self.possible_discounts = possible_discounts
        self.example_seats_are_real = example_seats_are_real
        self.discount_code = discount_code
        self.discount_desc = discount_desc
        self.discount_subdata = discount_subdata
        if free_seat_blocks is None:
            free_seat_blocks = []
        self.free_seat_blocks = free_seat_blocks
        self.raw_contiguous_seats = raw_contiguous_seats
        self.raw_total_seats = raw_total_seats
        self.user_commission = user_commission
        self.gross_commission = gross_commission

        vars(self).update(kwargs)


class Seat(CoreObject):

    def __init__(
        self,
        full_id=None,
        col_id=None,
        row_id=None,
        separator=None,
        is_restricted_view=None,
        seat_text=None,
        barcode=None,
        **kwargs
    ):
        self.full_id = full_id
        self.col_id = col_id
        self.row_id = row_id
        self.separator = separator
        self.is_restricted_view = is_restricted_view
        self.seat_text = seat_text
        self.barcode = barcode

        vars(self).update(kwargs)


class SeatBlock(CoreObject):

    def __init__(
        self,
        seat_block_token,
        block_length,
        seats=None,
        **kwargs
    ):
        self.seat_block_token = seat_block_token
        self.block_length = block_length
        if seats is None:
            seats = []
        self.seats = seats

        vars(self).update(kwargs)


class DespatchMethod(CoreObject):

    def __init__(
        self,
        despatch_type,
        despatch_desc,
        despatch_cost,
        despatch_token=None,
        despatch_code=None,
        **kwargs
    ):
        self.despatch_type = despatch_type
        self.despatch_desc = despatch_desc
        self.despatch_cost = despatch_cost
        self.despatch_token = despatch_token
        self.despatch_code = despatch_code

        vars(self).update(kwargs)


class Country(CoreObject):

    def __init__(
        self,
        country_code,
        country_desc,
        **kwargs
    ):
        self.country_code = country_code
        self.country_desc = country_desc

        vars(self).update(kwargs)


class Currency(CoreObject):

    def __init__(
        self,
        currency_code,
        currency_number,
        currency_pre_symbol,
        currency_post_symbol,
        currency_factor=None,
        currency_places=None,
        **kwargs
    ):
        self.currency_code = currency_code
        self.currency_number = currency_number
        self.currency_pre_symbol = currency_pre_symbol
        self.currency_post_symbol = currency_post_symbol
        self.currency_factor = currency_factor
        self.currency_places = currency_places

        vars(self).update(kwargs)


class Discount(CoreObject):

    def __init__(
        self,
        surcharge,
        ticket_price=None,
        discount_token=None,
        discount_type=None,
        discount_desc=None,
        discount_code=None,
        seatprice=None,
        no_of_tickets=None,
        seats=None,
        number_available=None,
        raw_contiguous_seats=None,
        raw_total_seats=None,
        user_commission=None,
        gross_commission=None,
        **kwargs
    ):
        self.discount_token = discount_token
        self.ticket_price = ticket_price
        self.surcharge = surcharge
        self.discount_type = discount_type
        self.discount_desc = discount_desc
        self.discount_code = discount_code
        self.seatprice = seatprice
        self.no_of_tickets = no_of_tickets
        if seats is None:
            seats = []
        self.seats = seats
        self.number_available = number_available
        self.raw_contiguous_seats = raw_contiguous_seats
        self.raw_total_seats = raw_total_seats
        self.user_commission = user_commission
        self.gross_commission = gross_commission

        vars(self).update(kwargs)


class Order(CoreObject):

    def __init__(
        self,
        item_number,
        venue_desc,
        event_desc,
        despatch_desc,
        ticket_type_desc,
        total_seatprice,
        total_surcharge,
        total_no_of_tickets,
        order_token=None,
        currency=None,
        despatch_method=None,
        discounts=None,
        performance=None,
        event=None,
        backend_purchase_reference=None,
        requested_seats=None,
        seat_request_status=None,
        user_commission=None,
        gross_commission=None,
        **kwargs
    ):
        self.item_number = item_number
        self.venue_desc = venue_desc
        self.event_desc = event_desc
        self.despatch_desc = despatch_desc
        self.ticket_type_desc = ticket_type_desc
        self.total_seatprice = total_seatprice
        self.total_surcharge = total_surcharge
        self.total_no_of_tickets = total_no_of_tickets
        self.order_token = order_token
        self.currency = currency
        self.despatch_method = despatch_method
        if discounts is None:
            discounts = []
        self.discounts = discounts
        self.performance = performance
        self.event = event
        self.backend_purchase_reference = backend_purchase_reference
        if requested_seats is None:
            requested_seats = []
        self.requested_seats = requested_seats
        self.seat_request_status = seat_request_status
        self.user_commission = user_commission
        self.gross_commission = gross_commission

        vars(self).update(kwargs)


class Trolley(CoreObject):

    def __init__(
        self,
        trolley_order_count,
        trolley_bundle_count,
        bundles=None,
        purchase_result=None,
        transaction_id=None,
        purchase_error=None,
        **kwargs
    ):
        self.trolley_order_count = trolley_order_count
        self.trolley_bundle_count = trolley_bundle_count
        if bundles is None:
            bundles = []
        self.bundles = bundles
        self.purchase_result = purchase_result
        self.transaction_id = transaction_id
        self.purchase_error = purchase_error

        vars(self).update(kwargs)


class Bundle(CoreObject):

    def __init__(
        self,
        bundle_source_desc,
        bundle_source_code,
        bundle_order_count,
        bundle_total_seatprice,
        bundle_total_surcharge,
        bundle_total_despatch,
        bundle_total_cost=None,
        orders=None,
        currency=None,
        purchase_result=None,
        **kwargs
    ):
        self.bundle_source_desc = bundle_source_desc
        self.bundle_source_code = bundle_source_code
        self.bundle_order_count = bundle_order_count
        self.bundle_total_seatprice = bundle_total_seatprice
        self.bundle_total_surcharge = bundle_total_surcharge
        self.bundle_total_despatch = bundle_total_despatch
        self.bundle_total_cost = bundle_total_cost
        if orders is None:
            orders = []
        self.orders = orders
        self.currency = currency
        self.purchase_result = purchase_result

        vars(self).update(kwargs)


class PurchaseResult(CoreObject):

    def __init__(
        self,
        success,
        failure_reason=None,
        is_semi_credit=None,
        is_partial=None,
        failed_cv_two=None,
        failed_avs=None,
        failed_3d_secure=None,
        **kwargs
    ):
        self.success = success
        self.failure_reason = failure_reason
        self.is_semi_credit = is_semi_credit
        self.is_partial = is_partial
        self.failed_cv_two = failed_cv_two
        self.failed_avs = failed_avs
        self.failed_3d_secure = failed_3d_secure

        vars(self).update(kwargs)


class CostRange(CoreObject):

    def __init__(
        self,
        currency,
        max_combined=None,
        max_surcharge=None,
        max_seatprice=None,
        min_combined=None,
        min_surcharge=None,
        min_seatprice=None,
        best_value_offer=None,
        max_saving_offer=None,
        top_price_offer=None,
        no_singles_cost_range=None,
        quantity_options=None,
        **kwargs
    ):
        self.max_combined = max_combined
        self.max_surcharge = max_surcharge
        self.min_combined = min_combined
        self.max_seatprice = max_seatprice
        self.min_seatprice = min_seatprice
        self.min_surcharge = min_surcharge
        self.best_value_offer = best_value_offer
        self.max_saving_offer = max_saving_offer
        self.top_price_offer = top_price_offer
        self.currency = currency
        self.no_singles_cost_range = no_singles_cost_range
        self.quantity_options = quantity_options

        vars(self).update(kwargs)


class Review(CoreObject):

    def __init__(
        self,
        is_user_review=None,
        review_date_desc=None,
        review_time_desc=None,
        review_date_yyyymmdd=None,
        review_time_hhmmss=None,
        review_title=None,
        review_body=None,
        review_author=None,
        review_lang=None,
        star_rating=None,
        **kwargs
    ):

        self.is_user_review = is_user_review
        self.review_date_desc = review_date_desc
        self.review_time_desc = review_time_desc
        self.review_date_yyyymmdd = review_date_yyyymmdd
        self.review_time_hhmmss = review_time_hhmmss
        self.review_title = review_title
        self.review_body = review_body
        self.review_author = review_author
        self.review_lang = review_lang
        self.star_rating = star_rating

        vars(self).update(kwargs)


class Customer(CoreObject):

    def __init__(
        self,
        first_name,
        first_name_latin,
        last_name,
        last_name_latin,
        town,
        town_latin,
        home_phone,
        work_phone,
        country_code,
        initials=None,
        initials_latin=None,
        suffix=None,
        suffix_latin=None,
        addr_line_one=None,
        addr_line_one_latin=None,
        addr_line_two=None,
        addr_line_two_latin=None,
        county=None,
        county_latin=None,
        postcode=None,
        postcode_latin=None,
        country=None,
        country_latin=None,
        title=None,
        title_latin=None,
        email_addr=None,
        dp_supplier=None,
        dp_user=None,
        dp_world=None,
        agent_ref=None,
        **kwargs
    ):

        self.first_name = first_name
        self.first_name_latin = first_name_latin
        self.initials = initials
        self.initials_latin = initials_latin
        self.last_name = last_name
        self.last_name_latin = last_name_latin
        self.suffix = suffix
        self.suffix_latin = suffix_latin
        self.addr_line_one = addr_line_one
        self.addr_line_one_latin = addr_line_one_latin
        self.addr_line_two = addr_line_two
        self.addr_line_two_latin = addr_line_two_latin
        self.town = town
        self.town_latin = town_latin
        self.county = county
        self.county_latin = county_latin
        self.postcode = postcode
        self.postcode_latin = postcode_latin
        self.country_code = country_code
        self.country = country
        self.country_latin = country_latin
        self.title = title
        self.title_latin = title_latin
        self.email_addr = email_addr
        self.home_phone = home_phone
        self.work_phone = work_phone
        self.dp_supplier = dp_supplier
        self.dp_user = dp_user
        self.dp_world = dp_world
        self.agent_ref = agent_ref

        vars(self).update(kwargs)


class SalePage(CoreObject):

    def __init__(
        self,
        sale_page_type,
        sale_page_subtype,
        sale_page,
        **kwargs
    ):

        self.sale_page_type = sale_page_type
        self.sale_page_subtype = sale_page_subtype
        self.sale_page = sale_page

        vars(self).update(kwargs)


class SelfPrintHTMLPage(CoreObject):

    def __init__(
        self,
        page_url,
        item_number,
        complete_page_url=None,
        **kwargs
    ):

        self.page_url = page_url
        self.item_number = item_number
        self.complete_page_url = complete_page_url

        vars(self).update(kwargs)


class Commission(CoreObject):

    def __init__(
        self,
        amount_excluding_vat,
        amount_including_vat,
        commission_currency,
        **kwargs
    ):

        self.amount_excluding_vat = amount_excluding_vat
        self.amount_including_vat = amount_including_vat
        self.commission_currency = commission_currency

        vars(self).update(kwargs)


class StructuredInfoItem(CoreObject):

    def __init__(
        self,
        key,
        name=None,
        value=None,
        **kwargs
    ):

        self.key = key
        self.name = name
        self.value = value

        vars(self).update(kwargs)


class AvailDetail(CoreObject):

    def __init__(
        self,
        avail_currency,
        seatprice,
        surcharge,
        full_seatprice=None,
        full_surcharge=None,
        absolute_saving=None,
        percentage_saving=None,
        day_mask=None,
        available_dates=None,
        **kwargs
    ):

        self.currency = avail_currency
        self.seatprice = seatprice
        self.surcharge = surcharge
        self.full_seatprice = full_seatprice
        self.full_surcharge = full_surcharge
        self.absolute_saving = absolute_saving
        self.percentage_saving = percentage_saving
        self.day_mask = day_mask
        self.available_dates = available_dates

        vars(self).update(kwargs)



import order as order_objs
from pyticketswitch.util import (
    to_int_or_none, format_price_with_symbol,
    to_float_or_none, resolve_boolean,
    to_int_or_return
)
from base import InterfaceObject


class Bundle(InterfaceObject):
    """Object that represents a TSW bundle.

    A TSW bundle is a collection of orders for a particular supplier,
    so Trolleys are made up of Bundles containing Orders. Information
    about a TSW bundle is available with this object. Bundle objects
    should always be accessed from a Trolley/Reservation object, the
    constructor is intended for internal use only.
    """

    def __init__(
        self,
        core_bundle,
        self_print_urls=None,
        **settings
    ):
        self._core_bundle = core_bundle
        self._orders = None
        self._self_print_urls = self_print_urls

        super(Bundle, self).__init__(**settings)

    @property
    def source_description(self):
        """Supplier name."""
        return self._core_bundle.bundle_source_desc

    @property
    def source_code(self):
        """Supplier code."""
        return self._core_bundle.bundle_source_code

    @property
    def order_count(self):
        """Number of orders in this Bundle."""
        return to_int_or_none(
            self._core_bundle.bundle_order_count
        )

    @property
    def total_seatprice(self):
        """Formatted string value of the total seat price with currency
        symbol."""
        return format_price_with_symbol(
            self._core_bundle.bundle_total_seatprice,
            self._core_bundle.currency.currency_pre_symbol,
            self._core_bundle.currency.currency_post_symbol
        )

    @property
    def total_surcharge(self):
        """Formatted string value of the total surcharge cost with currency
        symbol."""
        return format_price_with_symbol(
            self._core_bundle.bundle_total_surcharge,
            self._core_bundle.currency.currency_pre_symbol,
            self._core_bundle.currency.currency_post_symbol
        )

    @property
    def total_despatch(self):
        """Formatted string value of the total despatch cost with currency
        symbol."""
        return format_price_with_symbol(
            self._core_bundle.bundle_total_despatch,
            self._core_bundle.currency.currency_pre_symbol,
            self._core_bundle.currency.currency_post_symbol
        )

    @property
    def total_cost(self):
        """Formatted string value of the total combined price with currency
        symbol."""
        return format_price_with_symbol(
            self._core_bundle.bundle_total_cost,
            self._core_bundle.currency.currency_pre_symbol,
            self._core_bundle.currency.currency_post_symbol
        )

    @property
    def total_cost_float(self):
        """Float value of the total combined price."""
        return to_float_or_none(
            self._core_bundle.bundle_total_cost
        )

    @property
    def orders(self):
        """List of Order objects in this Bundle."""
        if self._orders is None:
            orders = []

            for core_order in self._core_bundle.orders:

                currency = (
                    core_order.currency or
                    self._core_bundle.currency or
                    None
                )

                order = order_objs.Order(
                    core_order=core_order,
                    core_currency=currency,
                    **self._internal_settings()
                )

                if self._self_print_urls:
                    for s in self._self_print_urls:

                        if to_int_or_return(
                            s.item_number
                        ) == order.item_number:

                            order.self_print_url = s.complete_page_url
                            order.self_print_relative_url = s.page_url

                orders.append(order)

            self._orders = orders

        return self._orders

    @property
    def is_purchase_successful(self):
        """Boolean indicating if the bundle was purchased successfully

        Will return True if the purchase was successful,
        False if the purchase was not successful and
        None if no purchase was attempted.
        """
        is_success = None

        if self._core_bundle.purchase_result:

            is_success = resolve_boolean(
                self._core_bundle.purchase_result.success
            )

        return is_success

    @property
    def purchase_failure_reason(self):
        """Reason why purchase failed, None if not attempted."""
        reason = None

        if self.is_purchase_successful is False:
            reason = self._core_bundle.purchase_result.failure_reason

        return reason



try:
    import xml.etree.cElementTree as xml
except ImportError:
    import xml.etree.ElementTree as xml
import random
import string
import datetime

import settings

__all__ = (
    'create_xml_from_dict', 'create_dict_from_xml',
    'create_dict_from_xml_element',
    'random_string_generator',
    'format_price', 'format_price_with_symbol',
    'yyyymmdd_to_date', 'hhmmss_to_time',
    'date_to_yyyymmdd', 'time_to_hhmmss',
    'dates_in_range', 'auto_date_to_slug',
    'slug_to_auto_date', 'resolve_boolean',
    'to_int_or_none', 'to_int_or_return',
    'to_float_or_none', 'to_float_summed',
    'to_float_or_zero'
)


def create_xml_from_dict(root_name, arg_dict):
    """ Create XML from dictionary

    Creates a XML Element Object from a dictionary (with a given 'root')
    Dictionary must be flat or contain only dict or list

    """
    root = xml.Element(root_name)

    return _recur_xml_dict(root, arg_dict)


def _recur_xml_list(root, key, arg_list):
    for value in arg_list:
        if type(value) is dict:
            tmp = xml.SubElement(root, key)
            _recur_xml_dict(tmp, value)
        elif type(value) is list:
            tmp = xml.SubElement(root, key)
            _recur_xml_list(root, key, value)
        elif type(value) is bool:
            if value:
                tmp = xml.SubElement(root, key)
        else:
            tmp = xml.SubElement(root, key)
            try:
                tmp.text = unicode(value)
            except:
                tmp.text = value
    return root


def _recur_xml_dict(root, arg_dict):
    for (key, value) in arg_dict.iteritems():
        if type(value) is dict:
            tmp = xml.SubElement(root, key)
            _recur_xml_dict(tmp, value)
        elif type(value) is list:
            _recur_xml_list(root, key, value)
        elif type(value) is bool:
            if value:
                tmp = xml.SubElement(root, key)
        else:
            tmp = xml.SubElement(root, key)
            try:
                tmp.text = unicode(value)
            except:
                tmp.text = value
    return root


def create_dict_from_xml(xml_string):
    """ Create dictionary from XML

    Creates a dictionary from a string of XML (tag, text becomes key, value).
    Returned dictionary will consist of dictionaries and lists.

    """
    tree = xml.fromstring(xml_string)
    root = xml.Element('root')
    root.append(tree)

    return create_dict_from_xml_element(root)


def _add_xml_elem_to_dict(key, value, ret_dict):
    if key in ret_dict:
        if not isinstance(ret_dict[key], list):
            child_list = [ret_dict[key]]
            ret_dict[key] = child_list
        ret_dict[key].append(value)
    else:
        ret_dict[key] = value
    return ret_dict


def create_dict_from_xml_element(root):
    """

    Creates a dictionary from an XML Element object
    (tag, text becomes key, value).
    Returned dictionary will consist of dictionaries and lists.

    """
    ret_dict = {}

    for child in root:
        if len(list(child)) == 0:
            ret_dict = _add_xml_elem_to_dict(child.tag, child.text, ret_dict)
        else:
            ret_dict = _add_xml_elem_to_dict(
                child.tag, create_dict_from_xml_element(child), ret_dict
            )

    return ret_dict


def dict_ignore_nones(**kwargs):
    return dict((k, v) for k, v in kwargs.items() if v is not None)


def random_string_generator(
    size=20, chars=string.ascii_uppercase + string.digits
):
    return ''.join(random.choice(chars) for x in range(size))


def format_price(price_string):
    return '{:.2f}'.format(float(price_string))


def format_price_with_symbol(price_string, pre_symbol, post_symbol):
    price_and_symbol = [
        pre_symbol,
        format_price(price_string),
        post_symbol
    ]

    price_and_symbol = [x for x in price_and_symbol if x is not None]

    return ''.join(price_and_symbol)


def yyyymmdd_to_date(date_yyyymmdd):
    year = int(date_yyyymmdd[0:4])
    month = int(date_yyyymmdd[4:6])
    day = int(date_yyyymmdd[6:8])

    return datetime.date(year, month, day)


def hhmmss_to_time(time_hhmmss):
    hour = int(time_hhmmss[0:2])
    minute = int(time_hhmmss[2:4])
    sec = int(time_hhmmss[4:6])

    return datetime.time(hour, minute, sec)


def date_to_yyyymmdd(date):
    return date.strftime('%Y%m%d')


def date_to_yyyymmdd_or_none(date):
    if date is None:
        return None
    return date_to_yyyymmdd(date)


def time_to_hhmmss(time):
    return time.strftime('%H%M%S')


def dates_in_range(first_date, last_date):

    dates = []

    delta = last_date - first_date

    for i in range(delta.days + 1):
        dates.append(first_date + datetime.timedelta(days=i))

    return dates


def auto_date_to_slug(auto_date):
    if auto_date:
        slug = None

        for mapping in settings.DATE_SLUG_MAP:
            if auto_date == mapping['auto_date']:
                slug = mapping['slug']

    return slug


def slug_to_auto_date(slug):
    if slug:
        auto_date = None

        for mapping in settings.DATE_SLUG_MAP:
            if slug == mapping['slug']:
                auto_date = mapping['auto_date']

    return auto_date


def resolve_boolean(yes_no_string):
    if yes_no_string:
        if yes_no_string == 'yes':
            return True
        return False
    return None


def boolean_to_yes_no(boolean):
    if boolean is True:
        return 'yes'
    else:
        return 'no'


def to_int_or_none(int_string):
    try:
        return int(int_string)
    except (ValueError, TypeError):
        return None


def to_int_or_return(to_convert):
    try:
        return int(to_convert)
    except ValueError:
        return to_convert


def to_float_or_none(float_string):
    try:
        return float(float_string)
    except (ValueError, TypeError):
        return None


def to_float_summed(*float_strings):
    total = float(0)
    all_none = True
    for fs in float_strings:
        try:
            fl = float(fs)
            all_none = False
        except ValueError:
            fl = float(0)

        total = total + fl

    if all_none:
        return None

    return total


def to_float_or_zero(float_string):
    try:
        return float(float_string)
    except ValueError:
        return float(0)


def day_mask_to_bool_list(day_mask):
    """Converts a day bit mask (used in TSW to represent a boolean value for
    days of the week) into a list of 7 booleans. The first item in the list
    represents the first day of the week.
    """
    bit_str = '{0:07b}'.format(int(day_mask))
    bit_str_list = list(bit_str)
    # Convert each element to a binary integer then convert to bool
    # List is reversed so that the lowest significant bit represents the first
    # day of the week
    bool_list = [bool(int(x)) for x in reversed(bit_str_list)]
    return bool_list
